#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


import copy

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from dolfin.common.plotting import mesh2triang
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.transforms import Affine2D
from matplotlib.tri import Triangulation

from . import dolfin
from .complex import *
from .utils.helpers import project_iterative

colors = dict(
    red=(210 / 255, 95 / 255, 95 / 255), green=(69 / 255, 149 / 255, 125 / 255)
)

if "gyptis" not in plt.colormaps():
    matplotlib.colormaps.register(
        cmap=LinearSegmentedColormap.from_list(
            "gyptis", [colors["green"], (1, 1, 1), colors["red"]], N=100
        )
    )
if "gyptis_r" not in plt.colormaps():
    matplotlib.colormaps.register(
        cmap=LinearSegmentedColormap.from_list(
            "gyptis_r", [colors["red"], (1, 1, 1), colors["green"]], N=100
        )
    )

if "gyptis_white" not in plt.colormaps():
    matplotlib.colormaps.register(
        cmap=LinearSegmentedColormap.from_list(
            "gyptis_white", [(1, 1, 1), (1, 1, 1), (1, 1, 1)], N=100
        )
    )

if "gyptis_black" not in plt.colormaps():
    matplotlib.colormaps.register(
        cmap=LinearSegmentedColormap.from_list(
            "gyptis_black", [(0, 0, 0), (0, 0, 0), (0, 0, 0)], N=100
        )
    )


def get_boundaries(markers, domain=None, shift=(0, 0)):
    data = markers.array()
    triang = mesh2triang(markers.mesh())
    ids = np.unique(data) if domain is None else [domain]
    triangulations = []
    for id in ids:
        triang_ = copy.deepcopy(triang)
        triang_.set_mask(data != id)
        triangulations.append(triang_)

    sub_bnds = []
    for triangtest in triangulations:
        maskedTris = triangtest.get_masked_triangles()
        verts = np.stack((triangtest.x[maskedTris], triangtest.y[maskedTris]), axis=-1)
        all_vert = np.vstack(verts).T
        sub_triang = Triangulation(*all_vert)

        boundaries = []
        for i in range(len(sub_triang.triangles)):
            boundaries.extend(
                (
                    sub_triang.triangles[i, j],
                    sub_triang.triangles[i, (j + 1) % 3],
                )
                for j in range(3)
                if sub_triang.neighbors[i, j] < 0
            )
        boundaries = np.asarray(boundaries)

        bndpnts = (
            shift[0] + sub_triang.x[boundaries].T,
            shift[1] + sub_triang.y[boundaries].T,
        )

        sub_bnds.append(bndpnts)

    return sub_bnds


def plot_boundaries(markers, domain=None, shift=(0, 0), ax=None, **kwargs):
    sub_bnds = get_boundaries(markers, domain=domain, shift=shift)
    if "c" not in kwargs and "color" not in kwargs:
        kwargs["color"] = "k"
    if ax is None:
        ax = plt.gca()
    lines = []
    for bndpnts in sub_bnds:
        line = ax.plot(*bndpnts, **kwargs)
        lines.append(line)
    return lines


# def plot_subdomains(markers, alpha=0.3):
#     a = dolfin.plot(markers, cmap="binary", alpha=alpha, lw=0.00, edgecolor="face")
#     return a
#     # a.set_edgecolors((0.1, 0.2, 0.5, 0.))


def plot_markers(markers, subdomains, ax=None, geometry=None, colorbar=True, **kwargs):
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    if "cmap" not in kwargs:
        kwargs["cmap"] = "inferno"
    plt.sca(ax)
    n = len(subdomains)
    cmap_disc = plt.cm.get_cmap(kwargs["cmap"])
    bounds = np.array(list(subdomains.values()))
    srt = np.argsort(bounds)
    ids = np.array(list(subdomains.keys()))[srt]
    bounds = bounds[srt]
    bounds = np.hstack([bounds, [bounds[-1] + 1]])

    tt = [(bounds[i + 1] + bounds[i]) / 2 for i in range(n)]

    norm = matplotlib.colors.BoundaryNorm(bounds, cmap_disc.N)
    kwargs["cmap"] = cmap_disc
    p = dolfin.plot(markers, norm=norm, **kwargs)
    if colorbar:
        cbar = plt.colorbar(
            matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap_disc), ticks=tt, ax=ax
        )

        cbar.ax.set_yticklabels(ids)
        cbar.ax.tick_params(length=0)
    else:
        cbar = None
    kwargs.pop("cmap")
    # cbar = plt.colorbar(p,ticks=list(subdomains.keys()),cmap=cmap_disc) if colorbar else None

    if geometry:
        geometry.plot_subdomains(**kwargs)
    return p, cbar


def plot_subdomains(markers, **kwargs):
    return plot_boundaries(markers, **kwargs)


def plot_real(
    fplot,
    ax=None,
    geometry=None,
    proj_space=None,
    colorbar=True,
    orientation="vertical",
    **kwargs,
):
    proj = proj_space is not None
    if "cmap" not in kwargs:
        kwargs["cmap"] = "RdBu_r"
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    if proj:
        fplot = project_iterative(fplot, proj_space)
    plt.sca(ax)
    p = dolfin.plot(fplot, **kwargs)
    cbar = plt.colorbar(p, orientation=orientation) if colorbar else None
    kwargs.pop("cmap")
    if geometry:
        geometry.plot_subdomains(**kwargs)
    return p, cbar


def plot_complex(
    fplot,
    ax=None,
    geometry=None,
    proj_space=None,
    colorbar=True,
    orientation="vertical",
    **kwargs,
):
    if ax is None:
        fig, ax = plt.subplots(1, 2)
    P, C = [], []
    for a, f in zip(ax, [fplot.real, fplot.imag]):
        p, cbar = plot_real(
            f,
            ax=a,
            geometry=geometry,
            proj_space=proj_space,
            colorbar=colorbar,
            orientation=orientation,
            **kwargs,
        )
        P.append(p)
        C.append(cbar)
    return P, C


def plot(fplot, ax=None, geometry=None, proj_space=None, colorbar=True, **kwargs):
    returnfunc = plot_complex if iscomplex(fplot) else plot_real
    return returnfunc(fplot, ax, geometry, proj_space, colorbar, **kwargs)


def pause(interval):
    backend = plt.rcParams["backend"]
    if backend in matplotlib.rcsetup.interactive_bk:
        figManager = matplotlib._pylab_helpers.Gcf.get_active()
        if figManager is not None:
            canvas = figManager.canvas
            if canvas.figure.stale:
                canvas.draw()
            canvas.start_event_loop(interval)
            return


def check_plot_type(plottype, f):
    if plottype == "real":
        fplot = f.real
    elif plottype == "imag":
        fplot = f.imag
    elif plottype == "module":
        fplot = f.module
    elif plottype == "phase":
        fplot = f.phase
    else:
        raise (
            ValueError(
                f"wrong plot type {plottype}, choose between real, imag, module or phase"
            )
        )
    return fplot



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

"""
pkg_resources compatibility shim for Python 3.12+

This module provides a minimal pkg_resources API for legacy packages
that haven't migrated to importlib.metadata yet.

Copyright (c) 2024
License: MIT or BSD (choose based on your gyptis license)
"""

import sys
import warnings
from importlib.metadata import PackageNotFoundError, distributions, version


class Distribution:
    """Minimal Distribution class to mimic pkg_resources.Distribution"""

    def __init__(self, name):
        self.project_name = name
        self.key = name.lower()

        # Try to get version, handling both hyphen and underscore naming
        try:
            self.version = version(name)
        except PackageNotFoundError:
            try:
                self.version = version(name.replace("-", "_"))
            except PackageNotFoundError:
                try:
                    self.version = version(name.replace("_", "-"))
                except PackageNotFoundError:
                    warnings.warn(
                        f"Could not find version for package '{name}'. "
                        "Using '0.0.0' as fallback.",
                        RuntimeWarning,
                    )
                    self.version = "0.0.0"


def get_distribution(name):
    """
    Get distribution metadata for a package.

    Args:
        name: Package name (e.g., 'fenics-ufl', 'numpy')

    Returns:
        Distribution object with .version attribute
    """
    return Distribution(name)


def require(requirements):
    """
    Ensure packages are available (simplified version).

    Args:
        requirements: String or list of requirement specifications

    Returns:
        List of Distribution objects
    """
    if isinstance(requirements, str):
        requirements = [requirements]

    results = []
    for req in requirements:
        # Parse requirement (simplified - just gets package name)
        pkg_name = req.split()[0].split(">=")[0].split("==")[0].split("<")[0]
        results.append(get_distribution(pkg_name))

    return results


def working_set():
    """
    Get all installed distributions (simplified version).

    Returns:
        List of Distribution objects for all installed packages
    """
    return [Distribution(dist.name) for dist in distributions()]


# Inject this module as pkg_resources into sys.modules
sys.modules["pkg_resources"] = sys.modules[__name__]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from .source import *


def plane_wave_2d(wavelength, theta, phase=0, amplitude=1, degree=1, domain=None):
    """
    Compute a 2D plane wave.

    Parameters
    ----------
    wavelength : float
        The wavelength of the plane wave.
    theta : float
        The angle of incidence in radians.
    phase : float, optional
        The phase shift of the plane wave. Default is 0.
    amplitude : float, optional
        The amplitude of the plane wave. Default is 1.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : dolfin.cpp.mesh.Mesh, optional
        The mesh for the domain of definition of the function.

    Returns
    -------
    expr : Expression
        The 2D plane wave as a dolfin Expression.
    """

    k0 = 2 * np.pi / wavelength
    K = k0 * np.array((-np.sin(theta), -np.cos(theta)))
    K_ = sympyvector(sp.symbols("kx, ky, 0", real=True))
    expr = amplitude * sp.exp(1j * (K_.dot(X) + phase))
    return expression2complex_2d(expr, kx=K[0], ky=K[1], degree=degree, domain=domain)


def plane_wave_3d(
    wavelength, theta, phi, psi, phase=(0, 0, 0), amplitude=1, degree=1, domain=None
):
    """
    Compute a 3D plane wave.

    Parameters
    ----------
    wavelength : float
        The wavelength of the plane wave.
    theta : float
        The angle of incidence in radians.
    phi : float
        The azimuthal angle of incidence in radians.
    psi : float
        The polarization angle of incidence in radians.
    phase : tuple of float, optional
        The phase shifts of the x, y, and z components of the plane wave.
        Default is (0, 0, 0).
    amplitude : float, optional
        The amplitude of the plane wave. Default is 1.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : dolfin.cpp.mesh.Mesh, optional
        The mesh for the domain of definition of the function.

    Returns
    -------
    expr : Expression
        The 3D plane wave as a dolfin Expression.
    """
    cx = np.cos(psi) * np.cos(theta) * np.cos(phi) - np.sin(psi) * np.sin(phi)
    cy = np.cos(psi) * np.cos(theta) * np.sin(phi) + np.sin(psi) * np.cos(phi)
    cz = -np.cos(psi) * np.sin(theta)

    c = np.array([cx, cy, cz])
    phase_shifts = np.array([np.exp((phis)) for phis in phase])
    Carray = c * phase_shifts
    C = dolfin.as_tensor([df.Constant(_c) for _c in Carray])

    k0 = 2 * np.pi / wavelength
    K = k0 * np.array(
        (
            -np.sin(theta) * np.cos(phi),
            -np.sin(theta) * np.sin(phi),
            -np.cos(theta),
        )
    )
    K_ = sympyvector(sp.symbols("kx, ky, kz", real=True))

    Propp = amplitude * sp.exp(1j * (K_.dot(X)))

    code = [sp.printing.ccode(p) for p in Propp.as_real_imag()]
    prop = dolfin.Expression(
        code, kx=K[0], ky=K[1], kz=K[2], degree=degree, domain=domain
    )

    return Complex(prop[0] * C, prop[1] * C)


class PlaneWave(Source):
    """
    PlaneWave class.

    Parameters
    ----------
    wavelength : float
        The wavelength of the plane wave.
    angle : float
        The angle of incidence in radians.
    dim : int, optional
        The dimension of the plane wave. Default is 2.
    phase : float or list of float, optional
        The phase shift of the plane wave. Default is 0.
        If dim is 3, phase can be a list of three floats.
    amplitude : float, optional
        The amplitude of the plane wave. Default is 1.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : dolfin.cpp.mesh.Mesh, optional
        The mesh for the domain of definition of the function.

    Returns
    -------
    expr : Expression
        The plane wave as a dolfin Expression.
    """

    def __init__(
        self, wavelength, angle, dim=2, phase=0, amplitude=1, degree=1, domain=None
    ):
        if dim == 3 and np.isscalar(phase):
            phase = (phase, phase, phase)
        super().__init__(
            wavelength,
            dim,
            phase=phase,
            amplitude=amplitude,
            degree=degree,
            domain=domain,
        )
        self.angle = angle

    @property
    def expression(self):
        """
        The plane wave as a dolfin Expression.

        Returns
        -------
        expr : Expression
            The plane wave as a dolfin Expression.
        """

        return (
            plane_wave_2d(
                self.wavelength,
                self.angle,
                phase=self.phase,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            )
            if self.dim == 2
            else plane_wave_3d(
                self.wavelength,
                *self.angle,
                phase=self.phase,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

from .ls import *


def grad_green_function_2d(
    wavelength, xs, ys, phase=0, amplitude=1, degree=1, domain=None
):
    """
    Compute the gradient of the 2D Green function associated with a point source.

    Parameters
    ----------
    wavelength : float
        The wavelength of the Green function.
    xs : float
        The x-coordinate of the point source.
    ys : float
        The y-coordinate of the point source.
    phase : float, optional
        The phase shift of the Green function. Default is 0.
    amplitude : float, optional
        The amplitude of the Green function. Default is 1.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : dolfin.cpp.mesh.Mesh, optional
        The mesh for the domain of definition of the function.

    Returns
    -------
    expr : Expression
        The gradient of the Green function as a dolfin Expression.
    """

    Xs = sympyvector(sp.symbols("xs, ys, 0", real=True))
    k0 = sp.symbols("k0", real=True)
    Xshift = X - Xs
    rho = sp.sqrt(Xshift.dot(Xshift))
    rho = rho.subs(x[2], 0)
    kr = k0 * rho
    k0_ = 2 * np.pi / wavelength
    KR = dolfin.Expression(
        sp.printing.ccode(kr), k0=k0_, xs=xs, ys=ys, degree=degree, domain=domain
    )
    R = dolfin.Expression(
        sp.printing.ccode(rho), xs=xs, ys=ys, degree=degree, domain=domain
    )
    Xshift_list = list(Xshift.components.values())

    re = dolfin.Expression("y1(KR)", KR=KR, degree=degree, domain=domain)
    im = dolfin.Expression("j1(KR)", KR=KR, degree=degree, domain=domain)
    A = (
        1
        / 4
        * Complex(re, im)
        * Constant(amplitude * k0_)
        / R
        * phase_shift_constant(ConstantRe(phase))
    )
    dg = [
        A
        * dolfin.Expression(
            sp.printing.ccode(coord), xs=xs, ys=ys, degree=degree, domain=domain
        )
        for coord in Xshift_list[:2]
    ]
    re = as_vector([_f.real for _f in dg])
    im = as_vector([_f.imag for _f in dg])
    out = Complex(re, im)
    return out


class Dipole(Source):
    """
    Dipole class.

    This class defines a dipole source for 2D electromagnetic simulations.

    Parameters
    ----------
    wavelength : float
        The wavelength of the dipole source.
    position : tuple of float
        The (x, y) coordinates of the dipole position.
    angle : float, optional
        The orientation angle of the dipole in radians. Default is 0.
    dim : int, optional
        The spatial dimension of the source. Default is 2.
    phase : float, optional
        The phase shift of the source. Default is 0.
    amplitude : float, optional
        The amplitude of the source. Default is 1.
    degree : int, optional
        The degree of the output Expression. Default is 1.
    domain : dolfin.cpp.mesh.Mesh, optional
        The mesh for the domain of definition of the function.

    Raises
    ------
    NotImplementedError
        If the dimension `dim` is 3, since Dipole is not implemented in 3D.
    """

    def __init__(
        self,
        wavelength,
        position,
        angle=0,
        dim=2,
        phase=0,
        amplitude=1,
        degree=1,
        domain=None,
    ):
        if dim == 3:
            raise NotImplementedError("Dipole not implemented in 3D")
        super().__init__(
            wavelength,
            dim=dim,
            phase=phase,
            amplitude=amplitude,
            degree=degree,
            domain=domain,
        )
        self.position = position
        self.angle = angle

    @property
    def expression(self):
        """
        The expression of the dipole.

        The expression is given by the gradient of a line source
        multiplied by a unit vector in the direction of the dipole.

        Returns
        -------
        expr : Expression
            The expression of the dipole as a dolfin Expression.
        """
        dls = grad_green_function_2d(
            self.wavelength,
            self.position[0],
            self.position[1],
            self.phase,
            self.amplitude,
            self.degree,
            self.domain,
        )
        n = as_vector(
            [ConstantRe(-np.sin(self.angle)), ConstantRe(-np.cos(self.angle))]
        )

        ### old lazy implementation
        # ls = LineSource(
        #     self.wavelength,
        #     self.position,
        #     self.dim,
        #     self.phase,
        #     self.amplitude,
        #     self.degree + 1,
        #     self.domain,
        # )
        # dls = grad(ls.expression)
        return dot(dls, n) / Constant(1j * self.wavenumber)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


import gmsh

gmsh.initialize()

# Create geometries
box1 = gmsh.model.occ.addBox(0, 0, 0, 1, 1, 1)
box2 = gmsh.model.occ.addBox(0.5, 0, 0, 0.1, 0.1, 0.1)
# sphere = gmsh.model.occ.addSphere(1.5, 0.5, 0.5, 0.3)

# Fragment preserves mapping
all_vols = [(3, box1), (3, box2)]
out, out_map = gmsh.model.occ.fragment(all_vols, [])

gmsh.model.occ.synchronize()

# Assign physical groups based on out_map
names = ["box1", "box2"]

print("\nFragment results (out_map):")
for i, entities in enumerate(out_map):
    print(f"  {names[i]}: {entities}")

# # Strategy 1: Each original geometry gets its fragments (may overlap)
# for i, entities in enumerate(out_map):
#     if entities:
#         tags = [e[1] for e in entities]
#         phys_tag = i + 1
#         gmsh.model.addPhysicalGroup(3, tags, phys_tag)
#         gmsh.model.setPhysicalName(3, phys_tag, names[i])

# Strategy 2 (commented): Assign unique regions only
# Find which volumes appear in multiple out_map entries
all_tags = [e[1] for entities in out_map for e in entities]
unique_tags = [tag for tag in all_tags if all_tags.count(tag) == 1]
shared_tags = list(set([tag for tag in all_tags if all_tags.count(tag) > 1]))

for i, entities in enumerate(out_map):
    tags = [e[1] for e in entities if e[1] in unique_tags]
    if tags:
        gmsh.model.addPhysicalGroup(3, tags, i + 1)
        gmsh.model.setPhysicalName(3, i + 1, names[i])

if shared_tags:
    gmsh.model.addPhysicalGroup(3, shared_tags, 999)
    gmsh.model.setPhysicalName(3, 999, "overlap")

# Generate mesh
gmsh.model.mesh.generate(3)

# Show physical groups in GUI:
# Tools → Options → Mesh → Surface faces/Volume faces → Color by physical group
print("\nPhysical groups assigned:")
for i, name in enumerate(names):
    print(f"  {name} → Physical Volume {i+1}")

gmsh.fltk.run()
gmsh.finalize()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

import importlib
import os
import sys
import warnings
from math import e, pi

import petsc4py
from scipy.constants import c, epsilon_0, mu_0

# Auto-inject pkg_resources compatibility shim for Python 3.12+
if sys.version_info >= (3, 12):
    try:
        import pkg_resources
    except ModuleNotFoundError:
        # Import our compatibility shim
        # This will inject itself as 'pkg_resources' in sys.modules
        from . import _pkg_resources_compat  # noqa: F401


warnings.filterwarnings(
    "ignore",
    message="pkg_resources is deprecated",
    category=UserWarning,
)

# Set PETSc options before initializing PETSc
petsc4py.init(
    [
        "petsc_prealloc",
        "200",
        "ksp_type",
        "preonly",
        "pc_type",
        "lu",
        "pc_factor_mat_solver_type",
        "mumps",
        # "options_left",
        # " 0",
    ]
)


from .__about__ import __author__, __description__, __version__


def use_adjoint(use_adj):
    """Short summary.

    Parameters
    ----------
    use_adj : bool
        Adds automatic differentiation with dolfin-adjoint if True.

    """
    if use_adj:
        os.environ["GYPTIS_ADJOINT"] = "1"
    else:
        try:
            del os.environ["GYPTIS_ADJOINT"]
        except Exception:
            pass

    import sys

    import gyptis

    importlib.reload(gyptis)
    its = [s for s in sys.modules.items() if s[0].startswith("gyptis")]
    for k, v in its:
        importlib.reload(v)


if os.environ.get("GYPTIS_ADJOINT") is not None:
    import dolfin

    importlib.reload(dolfin)
    import dolfin_adjoint

    ADJOINT = True
    dolfin.__dict__.update(dolfin_adjoint.__dict__)
    dolfin.__spec__.name = "dolfin"
    del dolfin_adjoint
    # prevents AttributeError: module 'fenics_adjoint.types.function'
    # has no attribute 'function' when writing solution to files with <<
    dolfin.function.function = dolfin.function
else:
    try:
        del dolfin
    except Exception:
        pass
    import dolfin

    importlib.reload(dolfin)
    ADJOINT = False


from dolfin import MPI

dolfin.parameters["form_compiler"]["cpp_optimize"] = True
dolfin.parameters["form_compiler"]["cpp_optimize_flags"] = "-O2"
# dolfin.parameters["form_compiler"]["quadrature_degree"] = 5
dolfin.parameters["reorder_dofs_serial"] = False
dolfin.parameters["ghost_mode"] = "shared_facet"


from .api import *
from .complex import *
from .plot import *
from .utils import logger, set_log_level



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


"""
Introduction to electromagnetic scattering: tutorial
https://www.osapublishing.org/josaa/fulltext.cfm?uri=josaa-35-1-163&id=380136
"""

import numpy as np
import pytest


def test_scatterring3d(shared_datadir):
    from gyptis import c, dolfin, epsilon_0, mu_0
    from gyptis.complex import Complex, assemble, cross, curl, dot
    from gyptis.models.scattering3d import BoxPML3D, Scatt3D
    from gyptis.sources import PlaneWave

    dolfin.parameters["form_compiler"]["quadrature_degree"] = 2

    degree = 1
    pmesh = 1
    pmesh_scatt = 1 * pmesh
    eps_sphere = 4
    SCSN = []
    Gamma = np.linspace(0.1, 5, 100)
    Gamma = [2]
    R_sphere = 0.25
    for gamma in Gamma:
        circ = 2 * np.pi * R_sphere
        lambda0 = circ / gamma
        b = R_sphere * 3
        box_size = (b, b, b)
        pml_width = (lambda0, lambda0, lambda0)
        g = BoxPML3D(box_size=box_size, pml_width=pml_width)
        radius_cs_sphere = 0.8 * min(g.box_size) / 2
        box = g.box
        sphere = g.add_sphere(0, 0, 0, R_sphere)
        sphere_cross_sections = g.add_sphere(0, 0, 0, radius_cs_sphere)
        sphere, sphere_cross_sections, box = g.fragment(
            sphere, [sphere_cross_sections, box]
        )

        g.add_physical([box, sphere_cross_sections], "box")
        g.add_physical(sphere, "sphere")
        surf = g.get_boundaries(sphere_cross_sections, physical=False)[0]
        g.add_physical(surf, "calc", dim=2)
        smin = 1 * R_sphere / 2
        s = min(lambda0 / pmesh, smin)
        smin_pml = lambda0 / (0.66 * pmesh)
        for coord in ["x", "y", "z", "xy", "xz", "yz", "xyz"]:
            g.set_mesh_size({f"pml_{coord}_box": smin_pml})

        g.set_size(box, s)
        g.set_size(sphere_cross_sections, s)
        g.set_size(surf, s, dim=2)
        s = min(lambda0 / (eps_sphere**0.5 * pmesh_scatt), smin)
        g.set_size(sphere, s)
        g.build()

        epsilon = dict(sphere=eps_sphere, box=1)
        mu = dict(sphere=1, box=1)

        pw = PlaneWave(
            wavelength=lambda0, angle=(0, 0, 0), dim=3, domain=g.mesh, degree=degree
        )
        bcs = {}
        s = Scatt3D(
            g,
            epsilon,
            mu,
            pw,
            boundary_conditions=bcs,
            degree=degree,
        )

        s.solve()
        Z0 = np.sqrt(mu_0 / epsilon_0)
        S0 = 1 / (2 * Z0)
        n_out = g.unit_normal_vector
        Es = s.solution["diffracted"]
        inv_mu_coeff = s.coefficients[1].invert().as_subdomain()
        omega = s.source.pulsation
        Hs = inv_mu_coeff / Complex(0, dolfin.Constant(omega * mu_0)) * curl(Es)
        Ps = dolfin.Constant(0.5) * cross(Es, Hs.conj).real
        Ws = -assemble(dot(n_out, Ps)("+") * s.dS("calc"))
        Sigma_s = Ws / S0
        S_sphere = R_sphere**2 * np.pi
        Sigma_s_norm = Sigma_s / S_sphere
        SCSN.append(Sigma_s_norm)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

import sys

import gyptis
from gyptis.mesh import read_mesh

meshname = sys.argv[1]
domnum = int(sys.argv[2])
dim = int(sys.argv[3])
outpath = sys.argv[4]

mesh_object = read_mesh(meshname, dim=dim)
mesh = mesh_object["mesh"]
cell = "triangle" if dim == 2 else "tetra"
markers = mesh_object["markers"][cell]


submesh = gyptis.dolfin.SubMesh(mesh, markers, domnum)

with gyptis.dolfin.XDMFFile(submesh.mpi_comm(), outpath) as infile:
    infile.write(submesh)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from gyptis.api import BoxPML
from gyptis.complex import ComplexFunctionSpace
from gyptis.formulations import *
from gyptis.geometry import *
from gyptis.materials import *
from gyptis.sources import PlaneWave


def test_maxwell2d():
    degree = 2
    wavelength = 0.3
    pmesh = 8
    lmin = wavelength / pmesh

    geom = BoxPML(
        dim=2,
        box_size=(4 * wavelength, 4 * wavelength),
        pml_width=(wavelength, wavelength),
    )
    cyl = geom.add_circle(0, 0, 0, 0.2)
    cyl, box = geom.fragment(cyl, geom.box)
    geom.add_physical(box, "box")
    geom.add_physical(cyl, "cyl")
    [geom.set_size(pml, lmin) for pml in geom.pmls]
    geom.set_size("box", lmin)
    geom.set_size("cyl", lmin)
    geom.build()
    mesh = geom.mesh_object["mesh"]

    epsilon = dict(box=1, cyl=3)
    mu = dict(box=1, cyl=1)

    stretch = 1 - 1j
    pmlx = PML("x", stretch=stretch, matched_domain="box", applied_domain="pml_x_box")
    pmly = PML("y", stretch=stretch, matched_domain="box", applied_domain="pml_y_box")
    pmlxy = PML(
        "xy", stretch=stretch, matched_domain="box", applied_domain="pml_xy_box"
    )

    epsilon_coeff = Coefficient(epsilon, geometry=geom, pmls=[pmlx, pmly, pmlxy])
    mu_coeff = Coefficient(mu, geometry=geom, pmls=[pmlx, pmly, pmlxy])

    coefficients = epsilon_coeff, mu_coeff
    V = ComplexFunctionSpace(mesh, "CG", degree)

    pw = PlaneWave(wavelength=wavelength, angle=0, dim=2, domain=mesh, degree=degree)

    bcs = {}

    Maxwell2D(
        geom,
        coefficients,
        V,
        source=pw,
        boundary_conditions=bcs,
        source_domains="cyl",
        reference="box",
    )

    ###########################

    geom = BoxPML(
        dim=2,
        box_size=(4 * wavelength, 4 * wavelength),
        pml_width=(wavelength, wavelength),
    )
    cyl = geom.add_circle(0, 0, 0, 0.2)
    box = geom.cut(geom.box, cyl)
    geom.add_physical(box, "box")
    cyl_bnds = geom.get_boundaries("box")[-1]
    geom.add_physical(cyl_bnds, "cyl_bnds", dim=1)
    [geom.set_size(pml, lmin) for pml in geom.pmls]
    geom.set_size("box", lmin)
    geom.build(0)
    mesh = geom.mesh_object["mesh"]
    epsilon = dict(box=1, cyl=3)
    mu = dict(box=1, cyl=1)

    stretch = 1 - 1j
    pmlx = PML("x", stretch=stretch, matched_domain="box", applied_domain="pml_x_box")
    pmly = PML("y", stretch=stretch, matched_domain="box", applied_domain="pml_y_box")
    pmlxy = PML(
        "xy", stretch=stretch, matched_domain="box", applied_domain="pml_xy_box"
    )

    epsilon_coeff = Coefficient(epsilon, geometry=geom, pmls=[pmlx, pmly, pmlxy])
    mu_coeff = Coefficient(mu, geometry=geom, pmls=[pmlx, pmly, pmlxy])

    coefficients = epsilon_coeff, mu_coeff
    function_space = ComplexFunctionSpace(mesh, "CG", degree)

    pw = PlaneWave(wavelength=wavelength, angle=0, dim=2, domain=mesh, degree=degree)

    bcs = {"cyl_bnds": "PEC"}

    maxwell = Maxwell2D(
        geom,
        coefficients,
        function_space,
        source=pw,
        boundary_conditions=bcs,
        source_domains="cyl",
        reference="box",
    )
    maxwell.build_boundary_conditions()


def test_maxwell2d_periodic():
    lambda0, period = 1, 1

    from gyptis.models.grating2d import Layered2D, OrderedDict
    from gyptis.sources.stack import make_stack

    thicknesses = OrderedDict(
        {
            "substrate": lambda0,
            "groove": lambda0,
            "superstrate": lambda0,
        }
    )

    degree = 1
    geom = Layered2D(period, thicknesses, lambda0)
    geom.build()
    epsilon = dict(
        {
            "substrate": 3,
            "groove": 1,
            "superstrate": 1,
        }
    )
    mu = dict(
        {
            "substrate": 1,
            "groove": 1,
            "superstrate": 1,
        }
    )
    stretch = 1 - 1j
    pml_top = PML(
        "y",
        stretch=stretch,
        matched_domain="superstrate",
        applied_domain="pml_y_superstrate",
    )
    pml_bottom = PML(
        "y",
        stretch=stretch,
        matched_domain="substrate",
        applied_domain="pml_y_substrate",
    )
    epsilon_coeff = Coefficient(epsilon, geometry=geom, pmls=[pml_top, pml_bottom])
    mu_coeff = Coefficient(mu, geometry=geom, pmls=[pml_top, pml_bottom])
    coefficients = epsilon_coeff, mu_coeff
    function_space = ComplexFunctionSpace(geom.mesh, "CG", degree)
    pw = PlaneWave(
        wavelength=lambda0, angle=np.pi / 4, dim=2, domain=geom.mesh, degree=degree
    )

    make_stack(
        geom,
        coefficients,
        pw,
        polarization="TM",
        source_domains=["groove"],
        degree=1,
        dim=2,
    )
    bcs = {}
    maxwell_per2d = Maxwell2DPeriodic(
        geom,
        coefficients,
        function_space,
        source=pw,
        boundary_conditions=bcs,
        source_domains=["groove"],
        reference="superstrate",
    )

    maxwell_per2d.build_lhs()
    maxwell_per2d.build_rhs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from . import dolfin
from .geometry import *


def prepare_boundary_conditions(bc_dict):
    valid_bcs = ["pec"]
    pec_bnds = []
    for bnd, cond in bc_dict.items():
        cond = cond.lower()
        if cond not in valid_bcs:
            raise ValueError(f"Unknown boundary condition {cond}")
        else:
            pec_bnds.append(bnd)
    return pec_bnds


def build_pec_boundary_conditions(pec_bnds, geometry, function_space, applied_function):
    boundary_conditions = []
    for bnd in pec_bnds:
        bc = DirichletBC(
            function_space,
            applied_function,
            geometry.boundary_markers,
            bnd,
            geometry.boundaries,
        )
        [boundary_conditions.append(b) for b in bc]
    return boundary_conditions


class _DirichletBC(dolfin.DirichletBC):
    def __init__(self, *args, **kwargs):
        self.subdomain_dict = args[-1]
        if not callable(args[2]):
            args = list(args)
            args[-2] = self.subdomain_dict[args[-2]]
            args = tuple(args[:-1])
        super().__init__(*args)


class DirichletBC:
    def __new__(cls, *args, **kwargs):
        W = args[0]
        value = args[1]
        Wre, Wim = W.split()
        bcre = _DirichletBC(Wre, value.real, *args[2:], **kwargs)
        bcim = _DirichletBC(Wim, value.imag, *args[2:], **kwargs)
        return bcre, bcim


class BiPeriodic2D(dolfin.SubDomain):
    def __init__(self, geometry, eps=dolfin.DOLFIN_EPS, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eps = eps
        self.vectors = geometry.vectors
        self.vertices = geometry.vertices

    def inside(self, x, on_boundary):
        on_bottom = is_on_line(x, self.vertices[0], self.vertices[1], eps=self.eps)
        on_left = is_on_line(x, self.vertices[3], self.vertices[0], eps=self.eps)

        on_vert_1 = dolfin.near(
            x[0], self.vertices[1][0], eps=self.eps
        ) and dolfin.near(x[1], self.vertices[1][1], eps=self.eps)
        on_vert_3 = dolfin.near(
            x[0], self.vertices[3][0], eps=self.eps
        ) and dolfin.near(x[1], self.vertices[3][1], eps=self.eps)
        return bool(
            (on_bottom or on_left) and not on_vert_1 and not on_vert_3 and on_boundary
        )

    def map(self, x, y):
        on_right = is_on_line(x, self.vertices[1], self.vertices[2], eps=self.eps)
        on_top = is_on_line(x, self.vertices[3], self.vertices[2], eps=self.eps)

        if on_right and on_top:
            y[0] = x[0] - self.vectors[0][0] - self.vectors[1][0]
            y[1] = x[1] - self.vectors[0][1] - self.vectors[1][1]
        elif on_right:
            y[0] = x[0] - self.vectors[0][0]
            y[1] = x[1] - self.vectors[0][1]
        elif on_top:
            y[0] = x[0] - self.vectors[1][0]
            y[1] = x[1] - self.vectors[1][1]
        else:
            y[0] = -10000
            y[1] = -10000


class PeriodicBoundary2DX(dolfin.SubDomain):
    def __init__(self, period, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.period = period

    def inside(self, x, on_boundary):
        return bool(dolfin.near(x[0], -self.period / 2) and on_boundary)

    def map(self, x, y):
        y[0] = x[0] - self.period
        y[1] = x[1]


class PeriodicBoundary3DZ(dolfin.SubDomain):
    def __init__(self, period, eps=dolfin.DOLFIN_EPS, map_tol=1e-10):
        self.period = period
        self.eps = eps
        self.map_tol = map_tol
        super().__init__(map_tol=map_tol)

    def near(self, x, y):
        return dolfin.near(x, y, eps=self.eps)

    def inside(self, x, on_boundary):
        return bool(self.near(x[0], -self.period / 2) and on_boundary)

    def map(self, x, y):
        y[0] = x[0]
        y[1] = x[1]
        y[2] = x[2] - self.period


class BiPeriodicBoundary3D(dolfin.SubDomain):
    def __init__(self, period, eps=dolfin.DOLFIN_EPS, map_tol=1e-10):
        self.period = period
        self.eps = eps
        self.map_tol = map_tol
        super().__init__(map_tol=map_tol)

    def near(self, x, y):
        return dolfin.near(x, y, eps=self.eps)

    def on_it(self, x, i, s="-"):
        s = -1 if s == "-" else 1
        return self.near(x[i], s * self.period[i] / 2)

    def inside(self, x, on_boundary):
        if on_boundary:
            if self.on_it(x, 0, "-") and not self.on_it(x, 1, "+"):
                return True
            return bool(self.on_it(x, 1, "-") and not self.on_it(x, 0, "+"))

    def map(self, x, y):
        y[0] = x[0] - self.period[0] if self.on_it(x, 0, "+") else x[0]
        y[1] = x[1] - self.period[1] if self.on_it(x, 1, "+") else x[1]
        y[2] = x[2]


class Periodic3D(dolfin.SubDomain):
    def __init__(self, geometry, eps=dolfin.DOLFIN_EPS, map_tol=1e-10):
        self.eps = eps
        self.map_tol = map_tol
        self.vectors = geometry.vectors
        self.vertices = geometry.vertices
        self.planes = geometry.planes
        super().__init__(map_tol=map_tol)

    def inside(self, x, on_boundary):
        ison0 = is_on_plane(x, *self.planes[4], eps=self.eps)
        ison1 = is_on_plane(x, *self.planes[2], eps=self.eps)
        ison2 = is_on_plane(x, *self.planes[0], eps=self.eps)

        ison_slave0 = is_on_line3D(x, self.planes[4], self.planes[2], eps=self.eps)
        ison_slave1 = is_on_line3D(x, self.planes[4], self.planes[0], eps=self.eps)
        ison_slave2 = is_on_line3D(x, self.planes[2], self.planes[0], eps=self.eps)

        return bool(
            (ison0 or ison1 or ison2)
            and (not ((ison_slave0) or (ison_slave1) or (ison_slave2)))
            and on_boundary
        )

    def map(self, x, y):
        v = self.vectors

        def set_coord(x, y):
            for i in range(3):
                y[i] = x[i]
            return y

        ison0 = is_on_plane(x, *self.planes[5], self.map_tol)
        ison1 = is_on_plane(x, *self.planes[3], self.map_tol)
        ison2 = is_on_plane(x, *self.planes[1], self.map_tol)
        if ison0 and ison1 and ison2:
            y = set_coord(x - v[0] - v[1] - v[2], y)

        elif ison0 and ison2:
            y = set_coord(x - v[0] - v[2], y)
        elif ison1 and ison2:
            y = set_coord(x - v[1] - v[2], y)
        elif ison0 and ison1:
            y = set_coord(x - v[0] - v[1], y)
        elif ison0:
            y = set_coord(x - v[0], y)
        elif ison1:
            y = set_coord(x - v[1], y)
        elif ison2:
            y = set_coord(x - v[2], y)
        else:
            y = set_coord([10000, 10000, 10000], y)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io

from ..formulations import Maxwell2D
from .twoscale2d import *


class _EigenProblemInclusion2D(Simulation):
    def __init__(
        self, geometry, epsilon=None, mu=None, boundary_conditions=None, degree=1
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        assert isinstance(geometry, Geometry)
        self.epsilon, self.mu = init_em_materials(geometry, epsilon, mu)
        function_space = ComplexFunctionSpace(geometry.mesh, "CG", degree)
        epsilon_coeff = Coefficient(self.epsilon, geometry, degree=degree)
        mu_coeff = Coefficient(self.mu, geometry, degree=degree)

        coefficients = epsilon_coeff, mu_coeff
        formulation = Maxwell2D(
            geometry,
            coefficients,
            function_space,
            degree=degree,
            polarization="TM",
            boundary_conditions=boundary_conditions,
            modal=True,
        )
        super().__init__(geometry, formulation)

        self.degree = degree


class HighContrastHomogenization2D(Simulation):
    def __init__(
        self,
        background,
        inclusion,
        epsilon=None,
        mu=None,
        degree=1,
        direction="x",
        direct=True,
        inclusion_boundaries="inclusion_bnds",
        boundary_conditions=None,
    ):
        assert isinstance(background, Lattice2D)
        assert isinstance(inclusion, Geometry)
        self.epsilon, self.mu = init_em_materials(background, epsilon, mu)
        epsilon_inclusion, mu_inclusion = init_em_materials(inclusion, epsilon, mu)
        self.epsilon.update(epsilon_inclusion)
        self.mu.update(mu_inclusion)
        self.background = background
        self.inclusion = inclusion
        # self.epsilon_background = dict(background=self.epsilon["background"])
        # self.mu_background = dict(background=self.mu["background"])

        self.epsilon_background = self.epsilon.copy()
        self.epsilon_background.pop("inclusion")
        self.mu_background = self.mu.copy()
        self.mu_background.pop("inclusion")

        self.epsilon_inclusion = self.epsilon.copy()
        self.epsilon_inclusion.pop("background")
        self.mu_inclusion = self.mu.copy()
        self.mu_inclusion.pop("background")

        self.hom2scale = Homogenization2D(
            self.background, self.epsilon_background, self.mu_background
        )

        boundary_conditions = boundary_conditions or {inclusion_boundaries: "PEC"}

        self.epb = _EigenProblemInclusion2D(
            self.inclusion,
            self.epsilon_inclusion,
            self.mu_inclusion,
            boundary_conditions=boundary_conditions,
            degree=degree,
        )

    def get_effective_permittivity(self, **kwargs):
        return self.hom2scale.get_effective_permittivity(**kwargs)

    def get_effective_permeability(self, k, neigs=10, target=0, tol=1e-12):
        self.eigs = self.epb.eigensolve(neigs, target, tol, half=False)
        Es = self.eigs["eigenvalues"]
        psis = self.eigs["eigenvectors"]
        mu_eff = 1
        dx = self.epb.dx
        for E, psi in zip(Es, psis):
            if not np.allclose(E, 1) and E.imag > 0:
                eps_ = self.epb.formulation.epsilon.as_subdomain()
                norm = assemble(eps_ * psi * psi.conj * dx)
                alpha = assemble(psi * dx)
                alpha1 = assemble(eps_ * psi.conj * dx)
                mu_eff += -(k**2 * 1) / (k**2 * 1 - E**2) * alpha * alpha1 / norm
        return mu_eff



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from .maxwell3d import *

zvec = Constant((0, 0, 1))


class MaxwellGuided3D(Maxwell3D):
    def __init__(self, *args, wavenumber=0, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.modal:
            raise NotImplementedError("Only modal analysis is implemented")
        self.wavenumber = wavenumber

        self._mu_inv = self.mu.invert().as_property(dim=3)
        self._epsilon = self.epsilon.as_property(dim=3)
        self._epsilon_dom = self.epsilon.as_subdomain()
        self._mu_inv_dom = self.mu.invert().as_subdomain()
        self.phasor = phasor(
            self.wavenumber,
            direction=2,
            degree=self.degree,
            domain=self.geometry.mesh,
        )

    def maxwell(self, u, v, epsilon, inv_mu, domain="everywhere"):
        epsilon = self._epsilon[domain]
        inv_mu = self._mu_inv[domain]
        k0 = Constant(self.wavenumber)
        if domain == []:
            return None
        # a₀(u,v) = ∫_Ω (∇ × v)·[μ̿⁻¹ (∇ × u)] dV
        form = [inner(inv_mu * curl(u), curl(v))]
        # a₁(u,v) = ∫_Ω {(∇ × v)·[μ̿⁻¹ (ẑ × u)] - (ẑ × v)·[μ̿⁻¹ (∇ × u)]} dV
        form.append(
            inner(inv_mu * cross(zvec, u), curl(v))
            - inner(cross(zvec, v), inv_mu * curl(u))
        )
        # a₂(u,v) = ∫_Ω (ẑ × v)·[μ̿⁻¹ (ẑ × u)] dV
        form.append(inner(inv_mu * cross(zvec, u), cross(zvec, v)))
        # m(e,v) = ∫_Ω v·[ε̿(r)u] dV
        form.append(-(k0**2) * inner(epsilon * u, v))
        return [f * self.dx(domain) for f in form]

    def _weak(self, u, v):
        formulation = 0
        for dom in self.geometry.domains.keys():
            form = self.maxwell(u, v, domain=dom)
            if form is None:
                continue
            else:
                formulation += form
        return [f.real + f.imag for f in formulation]

    @property
    def weak(self):
        return self._weak(self.trial, self.test)

    def build_boundary_conditions(self):
        self._boundary_conditions = self.build_pec_boundary_conditions(
            Constant((0, 0, 0))
        )
        return self._boundary_conditions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


import pytest
from test_geometry import geom2D

from gyptis import dolfin as df
from gyptis.optimize import *
from gyptis.plot import *
from gyptis.utils.helpers import array2function, rot_matrix_2d

np.random.seed(123456)

geom = geom2D(mesh_size=0.05)
mesh = geom.mesh_object["mesh"]
dx = geom.measure["dx"]
markers = geom.mesh_object["markers"]["triangle"]
domains = geom.subdomains["surfaces"]
r = geom.cyl_size


submesh = df.SubMesh(mesh, markers, domains["cyl"])

W = df.FunctionSpace(submesh, "DG", 0)
Wfilt = df.FunctionSpace(submesh, "CG", 2)
f = df.Expression(" sin(3*2*pi*(x[0]*x[1])/(r*r))", degree=0, r=r)
a0 = project(f, W)

a = array2function(np.random.rand(W.dim()), W)


def test_filter():
    af = filtering(a, 0)
    rfilt = r * 0.1
    af = filtering(a, rfilt)

    filter = Filter(rfilt)
    af = filter.apply(a)

    filter1 = Filter(rfilt, solver=filter.solver)
    filter1.apply(a)

    df.plot(af)
    rfilt_aniso = np.diag([0.458 * rfilt, 3 * rfilt])
    af_aniso = filtering(a, rfilt_aniso)

    df.plot(af_aniso)

    trot = np.pi / 3
    rot = rot_matrix_2d(trot)[:-1, :-1]
    rfilt_aniso_rot = rot.T @ rfilt_aniso @ rot
    af_aniso_rot = filtering(a, rfilt_aniso_rot)

    df.plot(af_aniso_rot)

    with pytest.raises(ValueError) as valerr:
        filtering(a, np.random.rand(5))
    assert "Wrong shape for rfilt" in str(valerr.value)


def test_simp(tol=1e-14):
    s_min, s_max, p = 4, 8, 1
    b = simp(a, s_min=s_min, s_max=s_max, p=p, complex=False)
    diff = project(b, W) - (s_min + (s_max - s_min) * a**p)
    err = assemble(abs(diff) * df.dx)
    assert err < tol
    s_min, s_max, p = 4 - 1j, 8 - 3j, 1
    np.random.seed(123456)
    A = np.random.rand(20)
    for ascalar in A:
        ascalar = 0.5
        b = simp(ascalar, s_min=s_min, s_max=s_max, p=p, complex=True)
        diff = b - (s_min + (s_max - s_min) * ascalar**p)
        err = abs(diff)
        assert err < tol


def test_projection():
    projection(a, beta=1, nu=0.5)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.2.0
# License: MIT
# See the documentation at gyptis.gitlab.io


from .maxwell3d import *


class Maxwell3DPeriodic(Maxwell3D):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        k0 = self.source.wavenumber
        theta0, phi0 = self.source.angle[:2]
        alpha0 = -k0 * np.sin(theta0) * np.cos(phi0)
        beta0 = -k0 * np.sin(theta0) * np.sin(phi0)
        gamma0 = -k0 * np.cos(theta0)
        self.propagation_vector = np.array([alpha0, beta0, gamma0])

        self.phasor_vect = [
            phasor(
                self.propagation_vector[i],
                direction=i,
                degree=self.degree,
                domain=self.geometry.mesh,
            )
            for i in range(3)
        ]
        self.phasor = self.phasor_vect[0] * self.phasor_vect[1]
        self.annex_field = make_stack(
            self.geometry,
            self.coefficients,
            self.source,
            source_domains=self.source_domains,
            degree=self.degree,
            dim=3,
        )

    @property
    def weak(self):
        u1 = (
            Constant((0, 0, 0))
            if self.modal
            else self.annex_field["as_subdomain"]["stack"]
        )
        u = self.trial * self.phasor
        v = self.test * self.phasor.conj
        return self._weak(u, v, u1)

    def build_boundary_conditions(self):
        applied_function = (
            Constant((0, 0, 0))
            if self.modal
            else -self.annex_field["as_subdomain"]["stack"] * self.phasor.conj
        )
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions
